/*
 *  SPL - The SPL Programming Language
 *  Copyright (C) 2004, 2005  Clifford Wolf <clifford@clifford.at>
 *
 *  This program is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  This program is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with this program; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 *
 *  builtins.c: Collection of some useful CLIB functions
 */

#define _GNU_SOURCE

#include <stdio.h>
#include <stdlib.h>
#include <math.h>

#include "spl.h"

/**
 * This module contains the SPL 'standard' builtin functions. It does not
 * need to be loaded explicitely, most SPL runtimes load it automatically
 * for you.
 */


/**
 * Write the texted passed as argument to standard ouput. Some SPL runtimes
 * replace this function with their own output function.
 */
// builtin write(text)
struct spl_node *spl_builtin_write(struct spl_task *task, void *data)
{
	char *text = spl_clib_get_string(task);
	printf("%s", text);
	fflush(stdout);
	return 0;
}

/**
 * Create a (unicode) charater from the number passed as argument.
 */
// builtin chr(number)
struct spl_node *spl_builtin_chr(struct spl_task *task, void *data)
{
	int number = spl_clib_get_int(task);
	char text[5] = { 0, 0, 0, 0, 0 };

#define OUT(bits, pos) ((number >> pos) & (0xff >> (8-bits)))

	if ( number <= 0x00007F ) {
		text[0] = number;
		return SPL_NEW_STRING_DUP(text);
	}

	if ( number <= 0x0007FF ) {
		text[0] = 0xC0 | OUT(5, 6);
		text[1] = 0x80 | OUT(6, 0);
		return SPL_NEW_STRING_DUP(text);
	}

	if ( number <= 0x00FFFF ) {
		text[0] = 0xE0 | OUT(4, 12);
		text[1] = 0x80 | OUT(6, 6);
		text[2] = 0x80 | OUT(6, 0);
		return SPL_NEW_STRING_DUP(text);
	}

	if ( number <= 0x10FFFF ) {
		text[0] = 0xF0 | OUT(3, 18);
		text[1] = 0x80 | OUT(6, 12);
		text[2] = 0x80 | OUT(6, 6);
		text[3] = 0x80 | OUT(6, 0);
		return SPL_NEW_STRING_DUP(text);
	}

#undef OUT

	/* Input is to big for beeing UTF-8 encoded. */
	return 0;
}

/**
 * Return the unicode character number of the first character in the text.
 */
// builtin ord(text)
struct spl_node *spl_builtin_ord(struct spl_task *task, void *data)
{
	char *t = spl_clib_get_string(task);

#define IN(off, mask, value) ((t[off] & (mask)) == (value))
#define OUT(off, bits, pos) ((t[off] & (0xff >> (8-bits))) << pos)

	if (IN(0, 0x80, 0x00))
		return SPL_NEW_INT(t[0]);

	if (IN(0, 0xE0, 0xC0) &&
	    IN(1, 0xC0, 0x80))
		return SPL_NEW_INT(OUT(0, 5, 6) | OUT(1, 6, 0));

	if (IN(0, 0xF0, 0xE0) &&
	    IN(1, 0xC0, 0x80) &&
	    IN(2, 0xC0, 0x80))
		return SPL_NEW_INT(OUT(0, 4, 12) |
				OUT(1, 6, 6) | OUT(2, 6, 0));

	if (IN(0, 0xF8, 0xF0) &&
	    IN(1, 0xC0, 0x80) &&
	    IN(2, 0xC0, 0x80) &&
	    IN(3, 0xC0, 0x80))
		return SPL_NEW_INT(OUT(0, 3, 18) |
				OUT(1, 6, 12) | OUT(2, 6, 6) | OUT(3, 6, 0));

#undef OUT
#undef IN

	/* Input is not UTF-8. This should not happen.. */
	return 0;
}

#define MATH1(x) spl_clib_reg(vm, #x, spl_builtin_math1_wrapper, x)
struct spl_node *spl_builtin_math1_wrapper(struct spl_task *task, void *data)
{
	double (*mlib_func)(double) = data;
	double arg1 = spl_clib_get_float(task);
	return SPL_NEW_FLOAT(mlib_func(arg1));
}

#define MATH2(x) spl_clib_reg(vm, #x, spl_builtin_math2_wrapper, x)
struct spl_node *spl_builtin_math2_wrapper(struct spl_task *task, void *data)
{
	double (*mlib_func)(double, double) = data;
	double arg1 = spl_clib_get_float(task);
	double arg2 = spl_clib_get_float(task);
	return SPL_NEW_FLOAT(mlib_func(arg1, arg2));
}

void spl_builtin_register_all(struct spl_vm *vm)
{
	spl_clib_reg(vm, "write", spl_builtin_write, 0);

	spl_clib_reg(vm, "chr", spl_builtin_chr, 0);
	spl_clib_reg(vm, "ord", spl_builtin_ord, 0);

	/**
	 * arc cosine function
	 */
	// builtin acos(x)
	MATH1(acos);

	/**
	 * arc sine function
	 */
	// builtin asin)
	MATH1(asin);

	/**
	 * arc tangent function
	 */
	// builtin atan(x)
	MATH1(atan);

	/**
	 * arc tangent function of two variables
	 */
	// builtin atan2(x, y)
	MATH2(atan2);

	/**
	 * cosine function
	 */
	// builtin cos(x)
	MATH1(cos);

	/**
	 * sine function
	 */
	// builtin sin(x)
	MATH1(sin);

	/**
	 * tangent function
	 */
	// builtin tan(x)
	MATH1(tan);

	/**
	 * square root function
	 */
	// builtin sqrt(x)
	MATH1(sqrt);

	/**
	 * round to nearest integer, away from zero
	 */
	// builtin round(x)
	MATH1(round);

	/**
	 * smallest integral value not less than argument
	 */
	// builtin ceil(x)
	MATH1(ceil);

	/**
	 * largest integral value not greater than argument
	 */
	// builtin floor(x)
	MATH1(floor);
}

